"""Setup developer API.

Settings:

.. autosummary::
   :toctree:

   UserSettings
   InstanceSettings
   StorageSettings

"""
from lamindb_setup.dev import (  # pragma: no cover
    InstanceSettings,
    StorageSettings,
    UserSettings,
)
